use strict;
use warnings;
use Test::More;

# generated by Dist::Zilla::Plugin::Test::PodSpelling 2.007005
use Test::Spelling 0.12;
use Pod::Wordlist;


add_stopwords(<DATA>);
all_pod_files_spelling_ok( qw( bin lib ) );
__DATA__
Andrew
Andrey
Anwar
Ashley
Ask
Autoincrementing
BSON
Bachtiar
Barr
Bernard
Binary
Bjørn
Bloor
Brendan
Brian
BulkWrite
BulkWriteResult
BulkWriteView
CN
Carey
Carvalho
Casey
ChangeStream
Chipman
Chodorow
Christian
ClientSession
Code
Colin
Collection
CommandResult
Cursor
Cyr
DBRef
DBRefs
DEPRECATIONS
Daniels
Danny
DataTypes
Database
David
DeleteResult
Denisov
Deprecations
Dhodapkar
Dirolf
DownloadStream
Eric
Error
Examples
FAILOVER
Fan
Filtered
Finn
Florian
Fowler
Friedman
GSSAPI
Gerard
Glenn
Golden
Goossen
Gorman
Graham
Gregoire
GridFS
GridFSBucket
Guimard
Hansen
Hao
Harish
Harnish
Ilmari
Inc
IndexView
InsertManyResult
InsertOneResult
Jackson
Jason
Johann
John
Joseph
Josh
Joshua
Juran
Kamil
Kempers
Ken
Khozov
Kimball
Kristina
Kunze
LDAP
Langner
Lee
Leich
Ltd
Mannsåker
Matt
Matthew
Matthews
Maurer
MaxKey
McAdams
MiB
Michael
Mike
MinKey
Mohammad
Mongo
MongoClient
MongoDB
Monitoring
Morrison
Moss
Multithreading
Nadle
Nickola
Nigel
Niko
Nuno
OID
Oberholtzer
ObjectId
Op
OpenSSL
Orlando
Othello
PCRE
Page
Pan
Pavel
QueryResult
RHEL
RTT
Raetzsch
Ragwitz
Rahul
Rassi
ReadConcern
ReadPreference
Regexp
Reis
ReplicaSetNoPrimary
ReplicaSetWithPrimary
Rezic
Robert
Robin
Rojas
Role
Rolschewski
Roman
Ronald
Rotmanov
Ryan
SASL
Sanbeg
Sedlacek
Shadowcat
Sharded
Shopsin
Slaven
Slowikowski
Steinbrunner
Stephen
Steve
Stewart
Storch
Stuart
Sturm
Systems
TCP
TLS
Tailable
Thomas
Timestamp
Tobias
Toffaletti
Trout
Trupcheff
Tutorial
Tyni
UnacknowledgedResult
UnknownTransactionCommitResult
Upadhyayula
UpdateResult
Upgrading
UploadStream
Uwe
Vazquez
Voelker
Walde
Wallace
Wan
Watt
Whitney
Williams
Willis
WriteConcern
Wu
Xavier
Xtreak
Yerin
Zhang
Zhihong
_Aggregate
_BatchInsert
_BulkWrite
_BypassValidation
_ChangeStream
_CollectionOp
_Command
_CommandCursorOp
_CommandMonitoring
_Constants
_Count
_CreateIndexes
_Credential
_CursorAPI
_DatabaseErrorThrower
_DatabaseOp
_Delete
_DeprecationWarner
_Dispatcher
_Distinct
_DropCollection
_DropDatabase
_DropIndexes
_EndTxn
_Explain
_FSyncUnlock
_FindAndDelete
_FindAndUpdate
_GetMore
_InsertOne
_InsertPreEncoder
_KillCursors
_Link
_ListCollections
_ListIndexes
_OpReplyParser
_ParallelScan
_Platform
_PrivateConstructor
_Protocol
_Query
_ReadOp
_ReadPrefModifier
_RenameCollection
_Server
_ServerSession
_SessionPool
_SessionSupport
_SingleBatchDocWrite
_Topology
_TopologyMonitoring
_TransactionOptions
_Types
_URI
_Update
_UpdatePreEncoder
_WriteOp
_WriteResult
addon
andrew
arrayFilter
arrayFilters
ashleyw
ask
avkhozov
bernard
bigjoe1008
brendan
bypassDocumentValidation
casey
causalConsistency
ccyr
cebjyre
chansen
cheeselee
chodorow
chunkSize
codec
coffeeshops
commandName
connectionId
contentType
copydbgetnonce
copydbsaslstart
createUser
danny
databaseName
david
diego
diegok
disambiguated
dmorrison
dropIndexes
dsteinbrunner
durationSecs
echowuhao
email
eric
errmsg
fallbacks
fh
friedo
fs
fsync
fullDocument
gbarr
geospatial
gerard
getnonce
gfs
guimard
hupadhyayula
ilmari
imsxlu
intra
ismaster
jak
jason
jjuran
joshua
journaling
jstewart
kallimachos
kb
keepalive
kid
kind
kslowikowski
kwilliams
langner
lib
libgsasl
listCollections
mapReduce
matt
maxScan
md
mestre
mike
mohammad
mongod
mongodb
mongos
mst
mulder
natively
newDescription
nigelg
nightsailer
nojournal
ns
ntyni
omaurer
onwards
oplog
ovazquez
pageview
pavel
phaylon
previousDescription
primaryPreferred
proxying
rafl
rahul
rassi
reconnections
refaddr
requestId
rescan
retryable
rkimball
rolschewski
rotmanov
roundtrip
ryan
saslContinue
saslStart
secondaryPreferred
sharded
sindbach
slaven
srezic
ssl
stevesanbeg
stevie
stuart
subdocument
tailable
tbsliver
tirkarthi
topologyId
toyou1995
trippable
trupcheff
trustno
updateUser
uploadDate
upsert
upserted
uwe
v1
walde
wallace
whjackson
writeConcern
wtimeout
zzh_621
