#  Copyright 2016 - present MongoDB, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

use strict;
use warnings;

use BSON;
use BSON::OID;

package TestCodecWrapper;

our @ISA = qw/BSON/;

sub create_oid {
    my $oid = BSON::OID->new();
    return bless $oid, "TestCodecWrapper::OID";
}

package TestCodecWrapper::OID;

our @ISA = qw/BSON::OID/;

1;

